import asyncio
import os
import secrets
import sys
from asyncio import FIRST_COMPLETED
from contextlib import redirect_stdout

import aioconsole
import click
import nacl.exceptions
import toml

import mzo
from mzo.utils import ENV_SETTER
from mzo.utils.authentication import test_access_token, ExpiredAccessToken, refresh_access_data
from mzo.utils.crypto import encrypt, decrypt
from mzo.utils.oauth_server import OAuthServer

OAUTH_APPLICATION_PROMPT = f"""\
Monzo currently have a limit on how many users a single developer can have using their \
applications (like this one) to 20 users.

To get around this you can instead register as a developer and become your own \
user. Here are the configuration details you will need:

             Name: Monzo CLI
         Logo URL: <blank>
    Redirect URLs: {mzo.OAUTH_REDIRECT_URI}
      Description: Command-line application for Monzo.
  Confidentiality: Confidential

Once completed you will be given a Client ID and a Client Secret which you can return here with.
"""

MONZO_OAUTH_CLIENT_CONSOLE_URL = "https://developers.monzo.com/apps/home"


MANUAL_BROWSER_PROMPT = """
    If your browser has not opened, please manually browse to this link:
    {url:}
"""


OAUTH_PROMPT = """\
Now that you've registered this application as your own you need to log in as your own first user.

Your Client ID and Secret will be saved so you will not need to provide this them again in future.
"""


SERVER_KILL_PROMPT = """\
Please complete the authentication process in the browser to grant this application access to \
your account.

Or hit [ENTER] to terminate this process.
"""

PASSWORD_PROMPT = """\
We want to make sure anyone using your machine can not just send themselves all your money, \
let's add a password. Keep it secret, keep it safe.
"""


@mzo.command(short_help='Authorization & session management')
@click.option('-f', '--format', 'fmt', type=click.Choice(['raw', 'cmd']), default='cmd', help='Chose the format the access token is return in.')
@click.option('--reauthorize', is_flag=True, help='Reauthorize the application\'s access to your Monzo account.')
@click.pass_context
async def login(ctx, reauthorize, fmt):
    """
    Manages authorizing this application to access your Monzo account as well as
    creating authenticated sessions.

    This function should be called within your shell's eval statement for the login session
    to be persisted.

    \b
    Authorization:

    The first time this command is called (or when it is subsequently called with the
    --reauthorize flag), the application request access to your Monzo account via OAuth.
    This will involve navigating you to the Monzo website to grant this application
    permissions to access your account by providing this application with a application-specific
    access token for your account.

    Once granted access, you will be prompted for a password to encrypt the access token
    so it can be stored securely on your machine and prevent anyone with access to your machine
    being able to have unfettered access to your Monzo account.

    \b
    Authenticated Sessions:

    After the application has been authorized to access your Monzo accounts calls to
    this command will give the ability to create a persisted login session within your
    terminal. This means that your password will not be needed for all commands executed
    within that session. This is done by decrypting the access token with your provided
    password and storing it in memory as a environment variable on the machine (aptly named:
    MZO_ACCESS_TOKEN).

    For convince this command will return the required shell command to set the environment
    variable (e.g. `export MZO_ACCESS_TOKEN="xxx"`) which means you can call this function
    within your shell's eval command so the above statement is executed for you and the
    environment variable set. However you may be using a shell that doesn't use the above syntax
    or want to manage the token yourself, in which case providing the `--format raw` will simply
    return the value of the access token instead of the shell command.
    """
    credentials_fp = os.path.join(ctx.obj.app_dir, 'credentials')
    have_credentials = os.path.isfile(credentials_fp)

    if reauthorize or not have_credentials:
        stderr_echo(OAUTH_APPLICATION_PROMPT)
        _ = click.prompt(
            text='Hit [ENTER] to open browser to developer console where you can register your own application',
            default='done',
            hide_input=True,
            show_default=False,
            err=True)

        click.launch(MONZO_OAUTH_CLIENT_CONSOLE_URL)
        stderr_echo(MANUAL_BROWSER_PROMPT.format(url=style_url(MONZO_OAUTH_CLIENT_CONSOLE_URL)))

        ctx.obj.client_id = click.prompt('Client ID', err=True)
        ctx.obj.client_secret = click.prompt('Client Secret', err=True)
        stderr_echo('\nPerfect!\n', color='green')

        access_data = await authorize(ctx)
        ctx.obj.access_token = access_data['access_token']
        default_account = await select_default_account(ctx)

        click.echo(PASSWORD_PROMPT, err=True)
        password = click.prompt("Password", err=True, confirmation_prompt=True, hide_input=True)
        encrypted_access_data = encrypt(toml.dumps(access_data).encode('utf-8'), password=password)

        os.makedirs(ctx.obj.app_dir, exist_ok=True)

        with open(os.path.join(ctx.obj.app_dir, 'credentials'), 'wb+') as fp:
            fp.write(encrypted_access_data)

        with open(os.path.join(ctx.obj.app_dir, 'config'), 'w+') as fp:
            toml.dump({
                'default': {
                    'account_id': default_account['id'],
                    'format': 'human',
                },
                'oauth': {
                    'client_id': ctx.obj.client_id,
                    'client_secret': ctx.obj.client_secret,
                }
            }, fp)

        stderr_echo(f'Client Authorized', color='green')
    else:
        with open(credentials_fp, 'rb') as fp:
            cipher_text = fp.read()

        while True:
            password = click.prompt("Password", hide_input=True, err=True)
            try:
                plain_text = decrypt(cipher_text, password=password)
            except nacl.exceptions.CryptoError:
                click.echo("Incorrect Password", err=True, color='red')
            else:
                access_data = toml.loads(plain_text.decode('utf-8'))

                try:
                    await test_access_token(access_data['access_token'], http_session=ctx.obj.http)
                except ExpiredAccessToken:
                    refresh_token = access_data['refresh_token']
                    access_data = await refresh_access_data(refresh_token, ctx=ctx)
                    encrypted_access_data = encrypt(toml.dumps(access_data).encode('utf-8'), password=password)

                    with open(credentials_fp, 'wb+') as fp:
                        fp.write(encrypted_access_data)
                finally:
                    break

    if fmt == 'raw':
        click.echo(access_data["access_token"])
    else:
        stderr_echo(f'Login Session Active', color='green')
        click.echo(ENV_SETTER.format(name='MZO_ACCESS_TOKEN', value=access_data["access_token"]))


def stderr_echo(message, color=None, underline=False):
    """ All output needs to be over stderr for prompts to show in eval. """
    message = click.style(message, fg=color, underline=underline)
    with redirect_stdout(sys.stderr):
        click.echo(message=message)


def style_url(url):
    return click.style(url, fg='blue', underline=True)


async def authorize(ctx):
    nonce = secrets.token_urlsafe(32)
    server = OAuthServer(
        client_id=ctx.obj.client_id,
        client_secret=ctx.obj.client_secret,
        nonce=nonce,
        http_session=ctx.obj.http)

    stderr_echo(OAUTH_PROMPT)
    _ = click.prompt(
        text='Hit [ENTER] to open your browser to authenticate this application to your Monzo account',
        default='done',
        hide_input=True,
        show_default=False,
        err=True)

    click.launch(server.auth_request_url)
    pretty_url = style_url(server.auth_request_url)
    stderr_echo(MANUAL_BROWSER_PROMPT.format(url=pretty_url) + '\n' + SERVER_KILL_PROMPT)

    await server.run()

    user_killed = asyncio.Task(aioconsole.ainput())
    got_access_token = asyncio.Task(server.access_token())

    completed, _ = await asyncio.wait([user_killed, got_access_token], return_when=FIRST_COMPLETED)

    if got_access_token in completed:
        return got_access_token.result()
    elif user_killed in completed:
        stderr_echo('Authentication Canceled', color='red')
        ctx.exit(0)
    else:
        stderr_echo('Error', color='red')
        ctx.exit(1)


async def select_default_account(ctx):
    url = 'https://api.monzo.com/accounts'
    headers = {'Authorization': f'Bearer {ctx.obj.access_token}'}

    resp = await ctx.obj.http.get(url, headers=headers)
    accounts = (await resp.json())['accounts']
    accounts = [a for a in accounts if not a['closed']]

    if len(accounts) > 1:
        raise NotImplementedError('cant handle multiple accounts currently')
    else:
        return accounts[0]



from collections import OrderedDict

import click

import mzo


@mzo.command(short_help='View all Monzo accounts.')
@click.pass_context
async def accounts(ctx):
    url = 'https://api.monzo.com/accounts'
    headers = {'Authorization': f'Bearer {ctx.obj.access_token}'}

    resp = await ctx.obj.http.get(url, headers=headers)
    accounts_ = (await resp.json())['accounts']

    choices = OrderedDict((a['id'], f"{a['description']}") for a in accounts_)




# -*- coding: utf-8 -*-
import click

from dateutil.parser import parse as iso8601_parse

import mzo.utils.authentication
from mzo.utils.formats import Format


@mzo.command(short_help='View account\'s current balance.')
@click.pass_context
@mzo.options.fmt()
@mzo.utils.authentication.authenticated
async def transactions(ctx, fmt: Format):
    params = {'account_id': ctx.obj.account_id, 'expand[]': 'merchant'}

    resp = await ctx.obj.http.get('https://api.monzo.com/transactions', params=params)
    payload = await resp.json()
    key_order = ['created', 'name', 'amount', 'category']
    transactions_ = [{
        'created': t['created'],
        'name': f"{t['merchant']['name']}" if t['merchant'] else t['description'],
        'amount': t['amount'] / 100,
        'category': t['category'].replace('_', ' ').title(),
    } for t in payload['transactions']]

    if fmt is Format.human:
        def fmt_row(key, value):
            if key == 'amount':
                if value and value > 0:
                    value = f'{value:+.2f}'
                    value = click.style(str(value), fg='green')
                elif value:
                    value = f'{abs(value):.2f}'
            if key == 'created':
                if value:
                    value = f'{iso8601_parse(value):%a %d %B}'

            return click.style(key.title(), bold=True), value

        transactions_ = (dict(fmt_row(k, v) for k, v in t.items()) for t in transactions_)
        key_order = [fmt_row(k, None)[0] for k in key_order]

    output = fmt.dumps(transactions_, keys=key_order)
    click.echo(output)




from functools import partial

import click

import mzo


name = partial(
    click.argument,
    'name',
    nargs=-1,
    metavar='<recipient-name>...',
    callback=lambda ctx, param, value: ' '.join(value),
)


amount = partial(
    click.argument,
    'amount',
    nargs=1,
    type=mzo.types.FloatRange(min=0),
    metavar='<amount>',
)




OAUTH_REDIRECT_URI = 'http://localhost:40004/oauth'


class ContextObject:

    def __init__(self, *, http, app_dir, client_id, client_secret, account_id, access_token):
        self.http = http
        self.app_dir = app_dir
        self.client_id = client_id
        self.client_secret = client_secret
        self.account_id = account_id
        self.access_token = access_token


from mzo.utils import group, command
from mzo import types, arguments, options
from mzo.balance import balance
from mzo.transactions import transactions
from mzo.login import login
from mzo.logout import logout



from asyncio import Event
from urllib.parse import urlencode

from sanic import Sanic
from sanic.exceptions import Unauthorized
from sanic.response import text

from mzo import OAUTH_REDIRECT_URI


class OAuthServer:

    def __init__(self, *, client_id, client_secret, nonce, http_session):
        self.http = http_session
        self.app = Sanic(__name__, configure_logging=False)
        self.client_id = client_id
        self.client_secret = client_secret
        self.nonce = nonce
        self._oauth_complete = Event()
        self._access_token = None

        @self.app.route('/favicon.ico')
        def ignore_favicon(_):
            return text('no favicon, sorry.')

        @self.app.route("/oauth")
        async def welcome_back(request):
            if request.args['state'][0] != self.nonce:
                raise Unauthorized("The nonce returned from Monzo does not match the one sent out.")

            resp = await self.http.post('https://api.monzo.com/oauth2/token', data={
                'grant_type': 'authorization_code',
                'client_id': client_id,
                'client_secret': client_secret,
                'redirect_uri': OAUTH_REDIRECT_URI,
                'code': request.args['code'][0]})
            self._access_token = await resp.json()
            self._oauth_complete.set()
            return text('Authenticated.')

    @property
    def auth_request_url(self):
        params = {
            'client_id': self.client_id,
            'redirect_uri': OAUTH_REDIRECT_URI,
            'response_type': 'code',
            'state': self.nonce}
        return f'https://auth.monzo.com?{urlencode(params)}'

    async def run(self):
        await self.app.create_server(host='localhost', port=40004, access_log=False)

    async def access_token(self):
        await self._oauth_complete.wait()
        return self._access_token



from functools import partial

import click

from mzo import types
from mzo.utils.formats import Format

fmt = partial(
    click.option,
    '--format',
    'fmt',
    metavar=f"({'|'.join(Format.__members__)})",
    type=types.Format,
    default=Format.human,
    help='The format the commands will output in.',
)

sort_code = partial(
    click.option,
    '-s',
    '--sort-code',
    metavar='<00-00-00>',
    help='Sort code of the recipient.',
)

account_number = partial(
    click.option,
    '-a',
    '--account-number',
    metavar='<00000000>',
    help='Account number of the recipient.',
)

message = partial(
    click.option,
    '-m',
    '--message',
    metavar='<message>',
    help='Message for recipient to attach to the payment.',
)

on_datetime = partial(
    click.option,
    '--on',
    metavar='<date/time>',
    default='now',
    help='When the payment should be made on or reoccur from.',
)

daily_frequency_flag = partial(
    click.option,
    '--daily',
    'every',
    flag_value=(1, 'days'),
    help='Schedule payment to reoccur daily.',
)

weekly_frequency_flag = partial(
    click.option,
    '--weekly',
    'every',
    flag_value=(1, 'weeks'),
    help='Schedule payment to reoccur weekly.',
)

monthly_frequency_flag = partial(
    click.option,
    '--monthly',
    'every',
    flag_value=(1, 'months'),
    help='Schedule payment to reoccur monthly.',
)

yearly_frequency_flag = partial(
    click.option,
    '--yearly',
    'every',
    flag_value=(1, 'years'),
    help='Schedule payment to reoccur annually.',
)


NEVER = (0, 'days')

user_defined_frequency = partial(
    click.option,
    '--every',
    'every',
    metavar='<x> (days|weeks|months|years)',
    type=(int, click.Choice(['days', 'weeks', 'months', 'years'])),
    default=NEVER,
    help='Schedule payment to reoccur at given interval.',
)



import asyncio
from collections import Awaitable
from functools import wraps

import click

from mzo.utils import csv
from mzo.utils.ascii_table import ascii_table

ENV_SETTER = """\
export {name:}="{value:}"
# This command is meant to be used with your shell's eval function.
# Run 'eval $(mzo login)' to sign into your Monzo account.
"""

NO_LOGIN_SESSION_ACTIVE = """\
No login session currently active. You can authorize this one-off command by providing your \
password, or see `mzo login --help` for persisting authentication between commands.
"""


def async_command(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)
        return wait(result)
    return wrapper


def command(name=None, cls=None, options_metavar='[options]', **attrs):
    def decorator(f):
        f = async_command(f)
        return click.command(name=name, cls=cls, options_metavar=options_metavar, **attrs)(f)
    return decorator


def group(name=None, options_metavar='[options]', **attrs):
    def decorator(f):
        f = async_command(f)
        return click.group(name=name, options_metavar=options_metavar, **attrs)(f)
    return decorator


def wait(f):
    if isinstance(f, Awaitable):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f)
    else:
        return f



from terminaltables import AsciiTable

from mzo.utils.dict_helpers import unify_dicts


def ascii_table(dicts=None, columns=None, title=None, fill='-'):
    if dicts:
        sorted_dicts = unify_dicts(dicts, key_order=columns, fill=fill)
        row_values = [d.values() for d in sorted_dicts]

        table = AsciiTable([[str(c) for c in columns]] + row_values, title=title)
        table.justify_columns = {i: 'right' if any(isinstance(r[c], (int, float)) for r in sorted_dicts) else 'left' for i, c in enumerate(columns)}

        return table.table
    else:
        return AsciiTable([['no results to display']], title=title or 'message').table



from collections.__init__ import OrderedDict

from ordered_set import OrderedSet


def unify_dicts(dicts, *, key_order=None, fill=None):
    if not key_order:
        key_order = OrderedSet(k for d in dicts for k in d.keys())

    filtered_dicts = [{k: v for k, v in d.items() if k in key_order}
                      for d in dicts]

    empty_template = {k: fill for k in key_order}
    filled_dicts = [{**empty_template, **d} for d in filtered_dicts]

    column_order = key_sorter(*key_order)
    sorted_dicts = [OrderedDict(sorted(d.items(), key=column_order))
                    for d in filled_dicts]

    return sorted_dicts


def key_sorter(*key_order):
    """
    Sort dictionary keys by provided order.

    This returns a function that cant be used as a key for a sort function,
    such as `fn` in `sorted(a, key=fn)`.
    """
    def column_index(dict_item):
        key, _ = dict_item
        try:
            return key_order.index(key)
        except ValueError:
            return float('infinity')
    return column_index



import asyncio
import os
from functools import partial

import aiohttp
import click
import toml
import uvloop

import mzo
from mzo.utils import wait
from mzo.__version__ import __version__

asyncio.set_event_loop(uvloop.new_event_loop())


@mzo.group()
@click.version_option(__version__)
@click.pass_context
async def cli(ctx):
    app_dir = click.get_app_dir('mzo', force_posix=True)
    config_fp = os.path.join(app_dir, 'config')

    client_id = None
    client_secret = None
    account_id = None
    access_token = os.environ.get('MZO_ACCESS_TOKEN')

    try:
        config = toml.load(config_fp)
    except FileNotFoundError:
        pass
    except toml.TomlDecodeError:
        click.echo(f'Unable to read config file "{config_fp}". Make sure it is valid TOML.')
        ctx.exit()
    else:
        oauth = config.get('oauth', {})
        client_id = oauth.get('client_id')
        client_secret = oauth.get('client_secret')

        default = config.get('default', {})
        account_id = default.get('account_id')

    headers = {'Authorization': f'Bearer {access_token}'} if access_token else None
    session = aiohttp.ClientSession(headers=headers)
    sync_close = partial(wait, session.close())
    ctx.call_on_close(sync_close)

    ctx.obj = mzo.ContextObject(
        http=session,
        app_dir=app_dir,
        client_id=client_id,
        client_secret=client_secret,
        account_id=account_id,
        access_token=access_token)


cli.add_command(mzo.login)
cli.add_command(mzo.logout)
cli.add_command(mzo.balance)

if os.environ.get('MZO_PRERELEASE') == '1':
    cli.add_command(mzo.transactions)


def main():
    cli()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())
    loop.close()



import sys

from contextlib import redirect_stdout

import click

import mzo

from mzo.utils import ENV_SETTER


@mzo.command(short_help='Logout of authenticated session.')
async def logout():
    click.echo(ENV_SETTER.format(name='MZO_ACCESS_TOKEN', value=""))

    with redirect_stdout(sys.stderr):
        message = click.style(f"Login Session Ended", fg='green')
        click.echo(message=message)



import os
from functools import wraps, partial

import aiohttp
import click
import nacl.exceptions
import toml

from mzo import OAUTH_REDIRECT_URI
from mzo.utils import NO_LOGIN_SESSION_ACTIVE, wait
from mzo.utils.crypto import decrypt, encrypt


class ExpiredAccessToken(Exception):
    ...


def authenticated(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        ctx = click.get_current_context()
        access_token = ctx.obj.access_token
        credentials_fp = os.path.join(ctx.obj.app_dir, 'credentials')
        have_credentials = os.path.exists(credentials_fp)

        # Test access token from active login session works
        if access_token:
            try:
                await test_access_token(access_token, http_session=ctx.obj.http)
            except ExpiredAccessToken:
                # fall back to trying to refresh - will require user password
                if have_credentials:
                    click.echo('Your access token has expired.'
                               'You will need to call `eval $(mzo login)` again to refresh '
                               'your token value.', err=True, color='red')
                    ctx.exit(1)
                else:
                    click.echo('Your access token had expired and there is no refresh token available. '
                           'Please reauthorize. See `mzo login --help`.', err=True, color='red')
                    raise click.Abort()

        # No session access token given, load from credentials file and test is it works
        elif have_credentials:
            click.echo(NO_LOGIN_SESSION_ACTIVE, err=True)

            with open(credentials_fp, 'rb') as fp_:
                cipher_text = fp_.read()

            password, access_data = retry_decrypt(cipher_text)

            try:
                await test_access_token(access_data['access_token'], http_session=ctx.obj.http)
            except ExpiredAccessToken:
                refresh_token = access_data['refresh_token']
                access_data = await refresh_access_data(refresh_token, ctx=ctx)
                encrypted_access_data = encrypt(toml.dumps(access_data).encode('utf-8'), password=password)

                with open(credentials_fp, 'wb+') as fp:
                    fp.write(encrypted_access_data)

            access_token = access_data['access_token']

        # No access token from session or from credentials. Point user in right direction
        else:
            click.echo('You are not logged in. See `mzo login --help`.', err=True, color='red')
            raise click.Abort()

        # Recreate session with any updates to access_token
        headers = {'Authorization': f'Bearer {access_token}'}
        session = aiohttp.ClientSession(headers=headers)
        sync_close = partial(wait, session.close())

        # update ctx
        ctx.call_on_close(sync_close)
        ctx.obj.http = session
        ctx.obj.access_token = access_token

        return await f(*args, **kwargs)
    return wrapper


def retry_decrypt(cipher_text):
    while True:
        # Keep this password for later if we need to refresh and restore new credentials
        password = click.prompt("Password", hide_input=True, err=True)
        try:
            plain_text = decrypt(cipher_text, password=password)
        except nacl.exceptions.CryptoError:
            click.echo("Incorrect Password", err=True, color='red')
        else:
            plain_text = toml.loads(plain_text.decode('utf-8'))
            return password, plain_text


async def test_access_token(access_token, *, http_session):
    resp = await http_session.get(
        url='https://api.monzo.com/ping/whoami',
        headers={'Authorization': f'Bearer {access_token}'})

    if resp.status == 401:
        raise ExpiredAccessToken()


async def refresh_access_data(refresh_token, *, ctx):
    client_id = ctx.obj.client_id
    client_secret = ctx.obj.client_secret

    if not client_id or not client_secret:
        click.Abort("Unable to find oauth id and secret in config file.")
    else:
        resp = await ctx.obj.http.post('https://api.monzo.com/oauth2/token', data={
            'grant_type': 'refresh_token',
            'client_id': client_id,
            'client_secret': client_secret,
            'redirect_uri': OAUTH_REDIRECT_URI,
            'refresh_token': refresh_token})

        payload = await resp.json()

        if 200 <= resp.status < 300:
            return payload
        else:
            click.Abort(payload['message'])



import json
from enum import Enum

from mzo.utils import ascii_table, csv
from mzo.utils.dict_helpers import unify_dicts


class Format(Enum):
    human = 'human'
    json = 'json'
    csv = 'csv'

    def dumps(self, dicts=None, keys=None, title=None, fill=None):
        if self is Format.human:
            return ascii_table(dicts=dicts, columns=keys, title=title, fill=fill)
        elif self is Format.json:
            dicts = unify_dicts(dicts, key_order=keys, fill=fill)
            return json.dumps(dicts, indent=2)
        elif self is Format.csv:
            dicts = unify_dicts(dicts, key_order=keys, fill=fill)
            return csv.dumps(dicts)
        else:
            raise NotImplementedError('Unknown output format', self)



import click

from mzo.utils.formats import Format as _Format


class FloatRange(click.ParamType):
    name = 'float range'

    def __init__(self, *, min=None, max=None, clamp=False):
        self.min_ = min
        self.max_ = max
        self.clamp = clamp

    def convert(self, value, param, ctx):
        try:
            value = float(value)
        except ValueError:
            self.fail('%s is not a valid float' % value, param, ctx)
        else:
            if self.min_ and value < self.min_ and self.clamp:
                return self.min_
            elif self.min_ and value < self.min_ and not self.clamp:
                self.fail(
                    '%s is smaller than the minimum valid value %s.'
                    % (value, self.min_), param, ctx)
            elif self.max_ and value > self.max_ and self.clamp:
                return self.max_
            elif self.max_ and value > self.max_ and not self.clamp:
                self.fail(
                    '%s is bigger than the maximum valid value %s.'
                    % (value, self.max_), param, ctx)
            else:
                return value


class FormatParamType(click.ParamType):
    name = 'format'

    def convert(self, value, param, ctx):
        try:
            value = value.lower()
        except AttributeError:
            pass
        finally:
            try:
                return _Format(value)
            except ValueError:
                self.fail('%s is not a valid format' % value, param, ctx)


Format = FormatParamType()



import nacl.encoding
import nacl.exceptions
import nacl.hash
import nacl.secret


def encrypt(data, *, password):
    secret_key = nacl.hash.sha256(password.encode('utf-8'), encoder=nacl.encoding.RawEncoder)
    secret_box = nacl.secret.SecretBox(secret_key)
    encrypted_data = secret_box.encrypt(data)
    del secret_key, secret_box
    return encrypted_data


def decrypt(data, *, password):
    secret_key = nacl.hash.sha256(password.encode('utf-8'), encoder=nacl.encoding.RawEncoder)
    secret_box = nacl.secret.SecretBox(secret_key)
    plain_text = secret_box.decrypt(data)
    del secret_key, secret_box
    return plain_text



import csv
import io

from ordered_set import OrderedSet


def dumps(dicts, header=True):
    output = io.StringIO()
    headers = OrderedSet(k for d in dicts for k in d.keys())

    writer = csv.DictWriter(output, fieldnames=headers)
    if header:
        writer.writeheader()
    writer.writerows(dicts)

    return output.getvalue()



# -*- coding: utf-8 -*-
import asyncio

import click

import mzo.utils.authentication
from mzo.utils.formats import Format


@mzo.command(short_help='View account\'s current balance.')
@click.pass_context
@mzo.options.fmt()
@mzo.utils.authentication.authenticated
async def balance(ctx, fmt: Format):
    params = {'account_id': ctx.obj.account_id}

    get_balance = ctx.obj.http.get('https://api.monzo.com/balance', params=params)
    get_pots = ctx.obj.http.get('https://api.monzo.com/pots', params=params)

    balance_resp, pots_resp = await asyncio.gather(get_balance, get_pots)
    balance_json = await balance_resp.json()
    pots_json = await pots_resp.json()

    def style_emoji(p):
        return {
            'cassette': '📼',
            'balls': '🎾',
            'beach_ball': '🏖',
            'rain': '☔',
            'fairy_lights': '💡',
            'yacht': '⛵',
            'piggy_bank': '🐷',
            'window': '🏠',
        }.get(p['style'], '🍯')

    rows = [
        {
            'name': 'Current Account',
            'balance': f'{balance_json["balance"]/100:.2f}',
            'emoji': '💸',
        },
        *[{
            'name': p['name'],
            'balance': f'{p["balance"]/100:.2f}',
            'emoji': style_emoji(p)
        } for p in pots_json['pots'] if not p['deleted']],
        {
            'name': 'Total',
            'balance': f'{balance_json["total_balance"]/100:.2f}',
            'emoji': '💰',
        }
    ]

    key_order = ['name', 'balance']

    if fmt is Format.human:
        def fmt_header(header):
            return click.style(header.title(), bold=True)

        key_order = [fmt_header(k) for k in key_order]

        rows = [{
            # replace row item's keys with header keys and concat emoji to name value
            fmt_header(k): f'{r["emoji"]} {v}' if k == 'name' else v for k, v in r.items()}
            for r in rows]

        # Spacer between total row
        rows.insert(-1, dict.fromkeys(key_order, ''))

    output = fmt.dumps(rows, keys=key_order)
    click.echo(output)
